/*
	PSEmu Plugin

	Plugin Type: GPU
	Plugin Name: Software Driver
	Description: All drawing is performed with software routines and DirectX3 is
				 used only to obtain pointer to surface.

	(C)1997-1998 Vision Thing Software Group

	All rights reserved. Using this source or part of it without permission
	is highly prohibited.

    Changes, bug reports - plugin@psemu.com
*/

#include "stdafx.h"
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <ddraw.h>
#define GPU_INTERNALS_DEF
#include "gpu_i.h"
#include "draw.h"
#include "resource.h"
#include "CnfDlg.h"
#include "blitter.h"
#include "primitive drawing.h"
#include "PSEmu PlugIn Defs.h"
#include "AbtDlg.h"
#include "syslog.h"

// PPDK developer must change libraryName field and can change revision and build

const unsigned char version = PLUGIN_VERSION;	// do not touch - library for PSEmu 1.x

// it is up to developer but values must be in range 0-255

const unsigned char revision =	1;
const unsigned char build =		9;

// to obtain library name for your plugin, mail: plugin@psemu.com
// this must be unique, and only we can provide this
static char *libraryName =		"Duddie's Soft Driver";


// driver dependant variables


// DirectX stuff follows here
static LPDIRECTDRAW            lpDD;           // DirectDraw object
static LPDIRECTDRAWSURFACE     lpDDSPrimary;   // DirectDraw primary surface
static LPDIRECTDRAWSURFACE     lpDDSPrimary2;   // DirectDraw primary surface
static DDSURFACEDESC       ddsd;
static DDSCAPS             ddscaps;
static HRESULT             ddrval;


GpuConfS gpuConfig;

typedef struct
{
	unsigned char bpp;
	unsigned short width;
	unsigned short height;
} GpuDispModeS;

unsigned long found32, found16;

static int ScreenOpened=0;
// internally used defines
static long initGPU = 0;

#define RED(x) (x & 0xff)
#define BLUE(x) ((x>>16) & 0xff)
#define GREEN(x) ((x>>8) & 0xff)

#define COLOR(x) (x & 0xffffff)

// macros for easy access to packet information
#define GPUCOMMAND(x) ((x>>24) & 0xff)

// memory image of the PSX vram 
unsigned char psxVub[1024*520*2];
signed char *psxVsb;
unsigned short *psxVuw;
signed short *psxVsw;
unsigned long *psxVul;
signed long *psxVsl;

long flip;

// internal GPU

static long GPUdataRet;
long GPUstatusRet;

static unsigned long gpuData[100];
static unsigned char gpuCommand = 0;
static long gpuDataC = 0;
static long gpuDataP = 0;

static long tWinX;
static long tWinY;
static long tWinW;
static long tWinH;

long drawingLines;

long dispWidths[4] = {256,320,512,640};
long dispPosX=0;
long dispPosY=0;
long dispVer = 0;
long dispHor = 0;
long dispHorNew;
long dispVerNew;
long dispColor = 0;
long dispColorNew;
long dispLace = 0;
long dispLaceNew;
long imageTransfer;
long drawLace;


short imTYc,imTXc,imTY,imTX;
long imSize;
short imageX0,imageX1;
short imageY0,imageY1;

unsigned short textBuf[512*512];
long newTextX0,newTextX1,newTextX2,newTextX3;
long newTextY0,newTextY1,newTextY2,newTextY3;

long gpuDataX;

void executePacket(void);






char * CALLBACK PSEgetLibName(void)
{
	return libraryName;
}

unsigned long CALLBACK PSEgetLibType(void)
{
	return  PSE_LT_GPU;
}

unsigned long CALLBACK PSEgetLibVersion(void)
{
	return version<<16|revision<<8|build;
}






void CALLBACK GPUmakeSnapshot(void)
{
 FILE *bmpfile;
 char filename[256];	 
 unsigned char header[0x36];
 long size;
 unsigned char line[1024*3];
 short i,j;
 unsigned char empty[2]={0,0};
 unsigned short color;
 unsigned long snapshotnr = 0;
 
 size=512*1024*3+0x38;
 
 // fill in proper values for BMP
 
 // hardcoded BMP header
 memset(header,0,0x36);
 header[0]='B';
 header[1]='M';
 header[2]=size&0xff;
 header[3]=(size>>8)&0xff;
 header[4]=(size>>16)&0xff;
 header[5]=(size>>24)&0xff;
 header[0x0a]=0x36;
 header[0x0e]=0x28;
 header[0x12]=1024%256;
 header[0x13]=1024/256;
 header[0x16]=512%256;
 header[0x17]=512/256;
 header[0x1a]=0x01;
 header[0x1c]=0x18;
 header[0x26]=0x12;
 header[0x27]=0x0B;
 header[0x2A]=0x12;
 header[0x2B]=0x0B;
 
 
 // increment snapshot value
 // get filename
 do
  {
   snapshotnr++;
   sprintf(filename,"SNAP\\ps%d.bmp",snapshotnr);
   bmpfile=fopen(filename,"rb");
   if (bmpfile == NULL) break;
   fclose(bmpfile);
  }
 while(TRUE);

 // try opening new snapshot file
 if((bmpfile=fopen(filename,"wb"))==NULL)
  {
   // could not open for writing !
   syslog(99,"PSE: Cannot snapshot to: [%s]\n",filename);
   return;
  }
 
 fwrite(header,0x36,1,bmpfile);
 for(i=511;i>=0;i--)
  {
   for(j=0;j<1024;j++)
    {
     color=psxVuw[i*1024+j];
     line[j*3+2]=(color<<3)&0xf1;
     line[j*3+1]=(color>>2)&0xf1;
     line[j*3+0]=(color>>7)&0xf1;
    }
   fwrite(line,1024*3,1,bmpfile);
  }
 fwrite(empty,0x2,1,bmpfile);
 fclose(bmpfile);  

}	    
 
HRESULT WINAPI enumModesCallback(LPDDSURFACEDESC s,
    LPVOID lpContext)
{
	if(s->ddpfPixelFormat.dwFlags & DDPF_RGB)
		if(s->dwWidth == 640)
			if (s->dwHeight == 480)
			{
				if(s->ddpfPixelFormat.dwRGBBitCount==32)
					found32=1;
				if(s->ddpfPixelFormat.dwRGBBitCount==16)
					found16=1;
			}

	return DDENUMRET_OK;
}

void DDInitFail(long ddrval);

long CALLBACK GPUinit()
{
	HKEY myKey;
	DWORD temp;
	DWORD type;
	DWORD size;

	ddrval = DirectDrawCreate( NULL, &lpDD, NULL );
	if( ddrval != DD_OK ) {DDInitFail(ddrval);return -1;}
	lpDD->EnumDisplayModes(0,NULL,NULL, enumModesCallback);
	lpDD->Release();
	if (found16 == 0 && found32 == 0)
		AfxMessageBox("Your DirectX driver do not support 16 and 32 bits per pixel color modes");

	if (RegOpenKeyEx(HKEY_CURRENT_USER,"Software\\Vision Thing\\PSEmu Pro\\GPU\\Soft",0,KEY_ALL_ACCESS,&myKey)==ERROR_SUCCESS)
	{
		size = 4;
		RegQueryValueEx(myKey,"H Lines",0,&type,(LPBYTE) &temp, &size);
		gpuConfig.scanH=(unsigned char)temp;
		size = 4;
		RegQueryValueEx(myKey,"V Lines",0,&type,(LPBYTE) &temp, &size);
		gpuConfig.scanV=(unsigned char)temp;
		RegCloseKey(myKey);
	}
	// initialise device here ... 
	// possibly configure the color format, etc.

	psxVsb=(signed char *)psxVub;
	psxVsw=(signed short *)psxVub;
	psxVsl=(signed long *)psxVub;
	psxVuw=(unsigned short *)psxVub;
	psxVul=(unsigned long *)psxVub;

	fillLookupTables();
	GPUstatusRet = 0x74000000;
	return 0;
}


long CALLBACK GPUshutdown()
{
	// shutdown device here....
	if(initGPU==1)
	{
		initGPU=0;
	}
	return 0;
}


static void finiObjects( void )
{
    if( lpDD != NULL )
    {
        if( lpDDSPrimary2 != NULL )
        {
            lpDDSPrimary2->Release();
            lpDDSPrimary2 = NULL;
        }
        if( lpDDSPrimary != NULL )
        {
            lpDDSPrimary->Release();
            lpDDSPrimary = NULL;
        }
		lpDD->Release();
		lpDD = NULL;
    }

} /* finiObjects */

void DDInitFail(long ddrval)
{
    finiObjects();
	syslog(0,"DirectDraw Init FAILED: %d",ddrval&0xffff);
    AfxMessageBox("Direct Draw init fail!", MB_OK ,0);
    return;
}


long CALLBACK GPUopen(HWND hwndGPU)
{
	ddrval = DirectDrawCreate( NULL, &lpDD, NULL );
	if( ddrval != DD_OK ) {DDInitFail(ddrval);return -1;}

    // Get exclusive mode
    ddrval = lpDD->SetCooperativeLevel( hwndGPU, DDSCL_EXCLUSIVE|DDSCL_FULLSCREEN);
    if( ddrval != DD_OK ) {DDInitFail(ddrval);return -1;}

    // Set the video mode to 640x480x16
	if (found16) ddrval = lpDD->SetDisplayMode( 640, 480, 16);
    if( ddrval != DD_OK )
		{DDInitFail(ddrval);return -1;}

    // Create the primary surface with 1 back buffer
    ddsd.dwSize = sizeof( ddsd );
	ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
	ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_FLIP | DDSCAPS_COMPLEX;
	ddsd.dwBackBufferCount = 1;
    ddrval = lpDD->CreateSurface( &ddsd, &lpDDSPrimary, NULL );
    if( ddrval != DD_OK ) {DDInitFail(ddrval);return -1;}

	ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
	ddrval = lpDDSPrimary->GetAttachedSurface(&ddscaps,&lpDDSPrimary2);
    if( ddrval != DD_OK ) {DDInitFail(ddrval);return -1;}

	DDBLTFX     ddbltfx;
    
	ddbltfx.dwSize = sizeof(ddbltfx);
	ddbltfx.dwFillColor = 0;
   
    ddrval = lpDDSPrimary->Blt(NULL,NULL,NULL,DDBLT_COLORFILL,&ddbltfx);
    ddrval = lpDDSPrimary2->Blt(NULL,NULL,NULL,DDBLT_COLORFILL,&ddbltfx);
      
	ScreenOpened=1;	
	return 0;
}

long CALLBACK GPUclose()
{
	// close the screen here
	// something like 'close()'
	if(ScreenOpened==1)
	{
//		snapshotDisplay();
        finiObjects();
		ScreenOpened=0;
	}
	return 0;
}


void updateDisplay(void)
{
 unsigned short * surf;
	HRESULT ddrval;
	__asm push edi
	ddrval=lpDDSPrimary2->Lock(NULL, &ddsd, DDLOCK_WAIT|DDLOCK_WRITEONLY, NULL);
	if(ddrval==DDERR_SURFACELOST)
	{
		syslog(99,"Restoring Surface\n");
		lpDDSPrimary2->Restore();
	}
	surf=(unsigned short *) (ddsd.lpSurface);

	if(ddrval!=DD_OK)
	{
		lpDDSPrimary2->Unlock(&ddsd);
		lpDDSPrimary->Flip(NULL,DDFLIP_WAIT);
		__asm pop edi
		return;
	}

//	if(gpuConfig.scanH==0 && gpuConfig.scanV==0)
	{
		//syslog(0,"Update START");
		if(dispVer==240)
		{
			if (found16)
				blitScreenMMX00s((unsigned long)surf, ddsd.lPitch-(dispHor*2), dispHor);
			else
				if (found32)
				blitScreen24_00((unsigned long)surf, ddsd.lPitch-(dispHor*4), dispHor);
		}
		else
		if(dispVer==480)
		{
			if (found16)
				blitScreenMMX00((unsigned long)surf, ddsd.lPitch-(dispHor*2), dispHor);
			else
				if (found32)
				blitScreen24_00((unsigned long)surf, ddsd.lPitch-(dispHor*4), dispHor);
		}
//		else
//			blitScreen24_01((unsigned long)surf, ddsd.lPitch-(dispHor*4), dispHor);

		//syslog(0,"Update END");
			//blitScreenMMX00s((unsigned long)surf, ddsd.lPitch-(dispHor*2), dispHor);
//		else
//		if(dispVer==480)
//				blitScreenMMX01((unsigned long)surf, ddsd.lPitch-(dispHor*2), dispHor);

/*
			if(dispHor==320)
			{
				n=dispPosY*1024+dispPosX;
				m=0;
				for(j=0;j<240;j++)
				{
					for(i=0;i<320;i++)
					{
						surf[m+i*2]=surf[m+i*2+1]=surf[m+(ddsd.lPitch/2)+i*2]=surf[m+(ddsd.lPitch/2)+i*2+1]=lutBGR2RGB[psxVuw[n+i]];
					}
					m+=(ddsd.lPitch);
					n+=1024;
				}
			}
			else
			if(dispHor==640)
			{
				n=dispPosY*1024+dispPosX;
				m=0;
				for(j=0;j<240;j++)
				{
					for(i=0;i<640;i++)
					{
						surf[m+i]=surf[m+(ddsd.lPitch/2)+i]=lutBGR2RGB[psxVuw[n+i]];
					}
					m+=(ddsd.lPitch);
					n+=1024;
				}
			}

*/

	}
/*	else
	if(gpuConfig.scanH==0 && gpuConfig.scanV==1)
	{
		if(dispVer==240)
		{
			if(dispHor==320)
			{
				n=dispPosY*1024+dispPosX;
				m=0;
				for(j=0;j<240;j++)
				{
					for(i=0;i<320;i++)
					{
						surf[m+i*2]=surf[m+(ddsd.lPitch/2)+i*2]=lutBGR2RGB[psxVuw[n+i]];
					}
					m+=(ddsd.lPitch);
					n+=1024;
				}
			}

		}
	}
	else
	if(gpuConfig.scanH==1 && gpuConfig.scanV==0)
	{
		if(dispVer==240)
		{
			if(dispHor==320)
			{
				n=dispPosY*1024+dispPosX;
				m=0;
				for(j=0;j<240;j++)
				{
					for(i=0;i<320;i++)
					{
						surf[m+i*2]=surf[m+i*2+1]=lutBGR2RGB[psxVuw[n+i]];
					}
					m+=(ddsd.lPitch);
					n+=1024;
				}
			}

		}
	}
	else
	if(dispVer==240)
	{
		if(dispHor==320)
		{
			n=dispPosY*1024+dispPosX;m=0;
			for(j=0;j<240;j++)
			{
				for(i=0;i<320;i++)
				{
					surf[m+i*2]=lutBGR2RGB[psxVuw[n+i]];
				}
				m+=(ddsd.lPitch);
				n+=1024;
			}
		}

	}
*/
	lpDDSPrimary2->Unlock(&ddsd);
	lpDDSPrimary->Flip(NULL,DDFLIP_WAIT);
	__asm pop edi
	__asm emms
	return;
}

void updateDisplayIfChanged(void)
{
 if ((dispVer == dispVerNew) && (dispHor == dispHorNew) && (dispColor == dispColorNew) && (dispLace == dispLaceNew)) return;

	DDBLTFX     ddbltfx;
    
	ddbltfx.dwSize = sizeof(ddbltfx);
	ddbltfx.dwFillColor = 0x000000;
   

	ddrval=lpDDSPrimary2->Lock(NULL, &ddsd, DDLOCK_WAIT|DDLOCK_WRITEONLY, NULL);
    ddrval = lpDDSPrimary->Blt(NULL,NULL,NULL,DDBLT_COLORFILL,&ddbltfx);
	lpDDSPrimary2->Unlock(&ddsd);

	lpDDSPrimary->Flip(NULL,DDFLIP_WAIT);

	ddrval=lpDDSPrimary2->Lock(NULL, &ddsd, DDLOCK_WAIT|DDLOCK_WRITEONLY, NULL);
    ddrval = lpDDSPrimary->Blt(NULL,NULL,NULL,DDBLT_COLORFILL,&ddbltfx);
	lpDDSPrimary2->Unlock(&ddsd);


	dispVer = dispVerNew;
	dispHor = dispHorNew;
	dispColor = dispColorNew;
	dispLace = dispLaceNew;
	updateDisplay();
}

// update lace is called evry VSync
void CALLBACK GPUupdateLace(void)
{
	__asm push edi
	if(dispLace)
	{
		drawingLines^=1;
		if(dispHor>0 && dispVer>0)
		{
			syslog(0,"UPDATING LACE");
			updateDisplay();
		}
	}
	__asm pop edi
}


// process read request from GPU status register
unsigned long CALLBACK GPUreadStatus(void)
{
	// return the status of the GPU
	/*
	- gp1 - GPU Status
          Mask
        04000000 - 1:Idle 0:Busy
        10000000 - 1:Ready 0:Not ready , to receive commands.
        80000000 - GPU is drawing 1:Odd 0:Even lines, in interlaced mode.
	*/
	//just'0x74000000' for now ! (idle, ready, even lines)
	if (drawingLines==1)
		return GPUstatusRet;
	else
		return GPUstatusRet|0x80000000;
	//return 0x74000000;
}

// processes data send to GPU status register
// these are always single packet commands.
void CALLBACK GPUwriteStatus(unsigned long gdata)
{
	switch((gdata>>24)&0xff)
	{
	case 0x04:
		syslog(77," GPU_S: TR_MODE: %x\n",gdata&0xffffff);
		if((gdata&0xffffff)==0) imageTransfer=0;
		if((gdata&0xffffff)==2) imageTransfer=3;
		return;
   
	case 0x05:
	    dispPosY = (gdata>>10)&0x3ff;
		dispPosX = gdata & 0x3ff;
		syslog(77," GPU_S: DISPLAY SET: X=%d Y=%d\n",dispPosX,dispPosY);
		if (dispLace==0) updateDisplay();
		return;

	case 0x08:
		dispHorNew = dispWidths[gdata&0x3];
		if(gdata&0x40 && dispHorNew == 320) dispHorNew=384;
		if(gdata&0x40 && dispHorNew == 256) dispHorNew=352;
		if (gdata&0x04) dispVerNew = 480;
		else dispVerNew = 240;
		dispColorNew = (gdata>>4)&0x1; // if 1 - TrueColor
		dispLaceNew = (gdata>>5)&0x01; // if 1 - Interlace
		syslog(77,"!GPU! DISPLAY SET: W=%d Wo=%d H=%d TRUE=%d LACE=%d\n",dispHorNew,dispWidths[gdata&0x3],dispVerNew,dispColorNew,dispLaceNew);
		if(dispVerNew==480 && dispLaceNew==1) drawLace = 1;
		else drawLace = 0;
		updateDisplayIfChanged();
		return;

	case 0x10: // ask about GPU version
		GPUdataRet=0x2; // set GPU version to 2
		return;

	default:
		syslog(77,"STATUS=%08x\n",gdata);
		return;
	}
	return;
}

unsigned long CALLBACK GPUreadData(void)
{
	if(imageTransfer==2)
	{
		// image transfer from VRAM
		if ((imTY>=0) && (imTY<512) && (imTX>=0) && (imTX<1024))
		{
			GPUdataRet=psxVul[imTY*512+imTX/2];
			syslog(11,"%08x",GPUdataRet);
		}
		imTX+=2;
		imTXc-=2;
		if(imTXc<=0)
		{
			imTX=imageX0;
			imTXc=imageX1;
			imTYc--;
			imTY++;
		}
		imSize--;          
		if(imSize <= 0)
		{
			GPUstatusRet&=0xf7ffffff;
			imageTransfer=0;
		}
	}
	return GPUdataRet;
}

// processes data send to GPU data register
// user should not take care about that
void CALLBACK GPUwriteData(unsigned long gdata)
{
	unsigned char command;
	GPUdataRet=gdata;
	//syslog(87,"%08x",gdata);
	if((imageTransfer & 1) == 1)
	{
		// image transfer to VRAM
		if (imTYc>0 && imTXc>0)
		{
			if ((imTY>=0) && (imTY<512) && (imTX>=0) && (imTX<1024))
				psxVuw[imTY*1024+imTX]=(unsigned short)gdata;
			imTX++;
			imTXc--;
			if(imTXc<=0)
			{
				imTX=(imageX0);
				imTXc=(imageX1);
				imTYc--;
				imTY++;
			}
			if(imTYc <= 0)
				imageTransfer=0;
     
			if ((imTY>=0) && (imTY<512) && (imTX>=0) && (imTX<1024))
				psxVuw[imTY*1024+imTX]=(unsigned short)(gdata>>16);

			imTX++;
			imTXc--;

			if(imTXc<=0)
			{
				imTX=(imageX0);
				imTXc=(imageX1);
				imTYc--;
				imTY++;
			}
     
			if(imTYc <= 0) 
				imageTransfer=0;
     
			return;
		}
		else
			imageTransfer=0;
	}
 
	//syslog(77,"!GPU! DATA = %08x\n",gdata);
	if(gpuDataC == 0)
	{
		command = (unsigned char) (gdata>>24) & 0xff;
		if (primTableC[command])
		{
			gpuDataC = primTableC[command];
			gpuCommand = command;
			gpuData[0] = gdata;
			gpuDataP = 1;
		}
		else 
		{
			if (gdata)
			{
				syslog(77,"!GPU! ERROR: COMMAND? %02x, %06x\n",command,gdata);
			}
			return;      
		}
	}
	else
	{
		gpuData[gpuDataP] = gdata;
		gpuDataP++;
	}
	if(gpuDataP == gpuDataC)
	{
		gpuDataC=gpuDataP=0;
		primTableJ[gpuCommand]((unsigned char *)gpuData);
	}
	return;
}

// this function will be removed soon
void CALLBACK GPUsetMode(unsigned long gdata)
{
	imageTransfer = gdata;
	return;
}

// this function will be removed soon
long CALLBACK GPUgetMode(void)
{
	return imageTransfer;
}


long CALLBACK GPUconfigure(void)
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState());
	CCnfDlg myDlg(NULL, &gpuConfig);
	myDlg.DoModal();
	return 0;
}

long CALLBACK GPUdmaChain(unsigned long * baseAddrL, unsigned long addr)
{
	// leave this function in peace :)
	// this is under high optimizing now
	// and however you dont need to change this
	
	unsigned long dmaMem;
	unsigned char * baseAddrB;
	unsigned char count;
	unsigned char cmd;

	__asm
	{
		push	esi
		push	edi
/*		mov		ebx,baseAddrL
		mov		eax,addr
		movzx	ecx, byte ptr[ebx+eax+3]	// count
dmaCountLoop:
		test	ecx,ecx
		jne		dmaCountLoop
*/
	}

	syslog(0,"*DMA CHAIN*");
	
	baseAddrB = (unsigned char*) baseAddrL;
	do
	{
		count = baseAddrB[addr+3];
		dmaMem = addr+4;
		while (count)
		{
			cmd = baseAddrB[dmaMem+3];
			//syslog(2,"     %08x %02x",baseAddrL[dmaMem/4],baseAddrB[dmaMem+3]);
			if (primTableC[cmd] == 0)
			{
				if (baseAddrL[dmaMem/4]!=0)
					syslog(255,"DMA_CMD *NI* %04x %04x",dmaMem,baseAddrL[dmaMem/4]);
				dmaMem+=4;
				count--;
			}
			else
			{
//				for (xx=0;xx<primTableC[cmd];xx++)
//					syslog(255,"DMA %04x %04x", dmaMem+xx*4,baseAddrL[dmaMem/4+xx]);
				primTableJ[cmd]((unsigned char*)baseAddrL + dmaMem);
				dmaMem+=(primTableC[cmd]*4);
				count-=primTableC[cmd];
			}
		}
		addr = baseAddrL[addr/4]&0xffffff;
		if (addr <= 0 || addr == 0xffffff) break;;
		addr&=0x7fffff;
	} while (TRUE);
	__asm
	{
		pop		edi
		pop		esi
	}
	// returns default
	return 0;
}



void CALLBACK GPUabout(void)
{
	// do not edit this function
	// modify only Resource Dialog (IDD_ABTDLG)
	AFX_MANAGE_STATE(AfxGetStaticModuleState());
	CAbtDlg myDlg;
	myDlg.DoModal();
}

long CALLBACK GPUtest(void)
{
	// if test fails this function should return negative value for error (unable to continue)
	// and positive value for warning (can continue but output might be crappy)
	return 0;
}
